#include "stdio.h"
#include "stdlib.h"
#include "string.h"

#include "stdarg.h"

#include "input.h"
#include "rc.h"
#include "loader.h"
#include "logger.h"

#include "Version"

extern emuquit;

static char *defaultconfig[] =
{
	"bind x quit",
	"bind 8 +up",
	"bind 2 +down",
	"bind 4 +left",
	"bind 6 +right",
	"bind a +a",
	"bind b +b",
	"bind p +start",
	"bind o +select",
	"bind u savestate",
	"bind i loadstate",
	"source gnuboy.rc",
	NULL
};


static void banner()
{
	/* printf("\ngnuboy " VERSION "\n");*/
}

static void copyright()
{
/*	banner();
	printf(
"Copyright (C) 2000-2001 Laguna and Gilgamesh\n"
"Portions contributed by other authors; see CREDITS for details.\n"
"\n"
"This program is free software; you can redistribute it and/or modify\n"
"it under the terms of the GNU General Public License as published by\n"
"the Free Software Foundation; either version 2 of the License, or\n"
"(at your option) any later version.\n"
"\n"
"This program is distributed in the hope that it will be useful,\n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"GNU General Public License for more details.\n"
"\n"
"You should have received a copy of the GNU General Public License\n"
"along with this program; if not, write to the Free Software\n"
"Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n"
"\n");
*/
}

static void usage(char *name)
{
	copyright();
	/* printf("Type %s --help for detailed help.\n\n", name); */
}

static void copying()
{
	copyright();
}

static void help(char *name)
{
	banner();
/*	printf("Usage: %s [options] romfile\n", name);
	printf("\n"
"      --source FILE             read rc commands from FILE\n"
"      --bind KEY COMMAND        bind KEY to perform COMMAND\n"
"      --VAR=VALUE               set rc variable VAR to VALUE\n"
"      --VAR                     set VAR to 1 (turn on boolean options)\n"
"      --no-VAR                  set VAR to 0 (turn off boolean options)\n"
"      --showvars                list all available rc variables\n"
"      --help                    display this help and exit\n"
"      --version                 output version information and exit\n"
"      --copying                 show copying permissions\n"
"");
*/
}


static void version(char *name)
{
/*	printf("%s-" VERSION "\n", name);*/
}


void doevents()
{
	event_t ev;
	int st;

	ev_poll();
	while (ev_getevent(&ev))
	{
		if (ev.type != EV_PRESS && ev.type != EV_RELEASE)
			continue;
		st = (ev.type != EV_RELEASE);
		rc_dokey(ev.code, st);
	}
}




static void shutdown()
{
	vid_close();
	pcm_close();
}

void die(char *fmt)
{
	Log(fmt);
	emuquit = 1;
	/*exit;*/
}

/*
static int bad_signals[] =
{

	SIGINT, SIGSEGV, SIGTERM, SIGFPE, SIGABRT, SIGILL,
#ifdef SIGQUIT
	SIGQUIT,
#endif
#ifdef SIGPIPE
	SIGPIPE,
#endif
	0
};

*/

static void fatalsignal(int s)
{
	die("Signal %d\n");
}

static void catch_signals()
{
/*	int i;
	for (i = 0; bad_signals[i]; i++)
		signal(bad_signals[i], fatalsignal);*/
}



static char *base(char *s)
{
	char *p;
	p = strrchr(s, '\\');
	if (p) return p+1;
	return s;
}

static char *basefolder(char *s)
{
	char *p,*ns;
	p = strrchr(s, '\\');
	if (p)
	{
	    ns = malloc(p-s+2);
        strncpy(ns,s,(p-s+1));
        return ns;
    }
	return "A:\\GAME\\";
}


int main(int argc, char** argv)
{
	int i;
	char *opt, *arg, *cmd, *s, *rom = 0, *var = 0, *var2 = 0;

	/* Avoid initializing video if we don't have to */
	/*for (i = 1; i < argc; i++)
	{
		if (!strcmp(argv[i], "--help"))
			help(base(argv[0]));
		else if (!strcmp(argv[i], "--version"))
			version(base(argv[0]));
		else if (!strcmp(argv[i], "--copying"))
			copying();
		else if (!strcmp(argv[i], "--bind")) i += 2;
		else if (!strcmp(argv[i], "--source")) i++;
		else if (!strcmp(argv[i], "--showvars"))
		{
			show_exports();
		}
		else if (argv[i][0] == '-' && argv[i][1] == '-');
		else if (argv[i][0] == '-' && argv[i][1]);
		else rom = "A:\\GAME\\SUPER2.GB";
	}*/

    rom = argv[0];
	if (!rom) usage(base(argv[0]));



    var = malloc(strlen("A:\\GAME\\")+1);
    strcpy(var,"A:\\GAME\\");
    rc_setvar("rcpath", 1, &var);
    free(var);

    var = basefolder(argv[0]);
	rc_setvar("savedir", 1, &var);
	init_exports();

	for (i = 0; defaultconfig[i]; i++)
		rc_command(defaultconfig[i]);


	cmd = malloc(strlen(rom) + 11);
	sprintf(cmd, "source %s", rom);
	s = strchr(cmd, '.');
	if (s) *s = 0;
	strcat(s, ".rc");
	rc_command(cmd);


    var = malloc(strlen(argv[0])+2);
    strcpy(var,argv[0]);
    s = strchr(var, '.');
    if (s) *s = 0;
    strcat(s, ".tga");

    var2 = basefolder(argv[0]);
    sprintf(var2,"%sgnuboy.tga",var2);

    vid_preinit();

	vid_init(var,var2);

	free(var);
    free(var2);

	pcm_init();


	loader_init(rom,basefolder(argv[0]));
	emu_reset();
	emu_run();



    cleanup();
    loader_unload();
    shutdown();

	return 0;
}











