/*
 * sdl.c
 * sdl interfaces -- based on svga.c
 *
 * (C) 2001 Damian Gryski <dgryski@uwaterloo.ca>
 * Joystick code contributed by David Lau
 * Sound code added by Laguna
 *
 * Licensed under the GPLv2, or later.
 */

#define MEMCPY newmemcpy


#include "../../lib/stdlib/stdlib.h"
#include "../../lib/stdlib/string.h"
#include "../../lib/stdlib/time.h"
#include "../../lib/stdlib/malloc.h"
#include "lib/graphics.h"
#include "lib/display.h"
#include "lib/control.h"
#include "lib/timer.h"
#include "lib/fixmath.h"
#include "../../lib/fsys.h"

#include "bigfontwhite.h"
#include "bigfontred.h"
#include "font.h"
#include "loading.h"
#include "frame.h"

#include "../../lib/audio.h"
#include "../../lib/ucos2.h"

#include "../../logger.h"
#include "fb.h"
#include "input.h"
#include "rc.h"

#define US(n) ( ((long)(n)) * 1000000 / CLOCKS_PER_SEC )

struct fb fb;

static int use_yuv = -1;
static int fullscreen = 1;
static int use_altenter = -1;
static int use_joy = -1;
static int fps = 0;
static int framecounter = 0;
static clock_t time1=0;
static int saveslot = 1;

extern short int emuquit;

static display* screen  = NULL;
static byte *fakescreen = NULL;

static gfx_texture *font = NULL;
static gfx_texture *bigfontred = NULL;
static gfx_texture *bigfontwhite = NULL;
static gfx_texture *loadingscreen = NULL;
static gfx_texture *frame = NULL;
static bool frameskip = false;
static bool useframeskip = false;
static int vmode[3] = { 0, 0, 16 };

static bool showfps = false;

rcvar_t vid_exports[] =
{
	RCV_END
};



rcvar_t joy_exports[] =
{
	RCV_END
};




void *sys_timer()
{
	clock_t *cl;

	cl = malloc(sizeof *cl);
	*cl = clock();
	return cl;
}

void sys_sleep(int us)
{
    clock_t start;
	if (us <= 0) return;
	start = clock();
	while(US(clock()-start) < us);
}


int sys_elapsed(clock_t *cl)
{
    clock_t now;
	int usecs;

	now = clock();
	usecs = US(now - *cl);
	*cl = now;
	return usecs;
}

#include "pcm.h"




struct pcm pcm={0, 0,0, NULL, 0};

#define A320SAMPLE_RATE 44100
#define STEREO 1
#define SNDLEN 4096*(STEREO+1)

static int sound = 1;
static int samplerate = A320SAMPLE_RATE;
static int stereo = STEREO;

static int volume = 25;

typedef void waveout_inst;

typedef struct {
   unsigned int sample_rate;
   unsigned short format;
   unsigned char channel;
   unsigned char volume;
} waveout_args;



waveout_inst *_audio_music = NULL;
waveout_args tempArgs = { A320SAMPLE_RATE, 16, STEREO+1,25};

#define  TASK_START_PRIO 5
#define  TASK_STK_SIZE   1024                      // Size of each task's stacks (# of WORDs)
#define  NO_TASKS        1                         // Number of identical tasks

OS_STK   TaskStk[NO_TASKS][TASK_STK_SIZE];          // Tasks stacks
OS_STK   TaskStartStk[TASK_STK_SIZE];
char     TaskData[NO_TASKS];                        // Parameters to pass to each task

static volatile int a320_sound_thread_exit=0;           // Flag to end a320_sound_thread_play() thread
static volatile int audio_done=1;
static volatile int a320_sound_paused = 0;


static byte a320_sound_buffer[SNDLEN];
static byte gb_sound_buffer[SNDLEN>>1];

rcvar_t pcm_exports[] =
{
	RCV_END
};


void a320_sound_thread_play(void *none)
{
    int teller = 0;
    int teller2 = 0;
    signed short w;
    do {
        if(a320_sound_paused == 0)
        {
            byte * bleh = (byte *) a320_sound_buffer;
            for (teller = 0; teller < SNDLEN>>1; teller++)
            {
                w = (uint16_t)((gb_sound_buffer[teller] - 128) << 8);
                *bleh++ = w & 0xFF ;
                *bleh++ = w >> 8;

            }

//            short *bleh = (short *) a320_sound_buffer;
//            for (teller = 0,teller2=0; teller < SNDLEN>>1; teller++,teller2+=2)
//            {
//               *bleh = (uint16_t)gb_sound_buffer[teller2]  | (uint16_t)gb_sound_buffer[teller2+1];
//            }
        waveout_write(_audio_music, a320_sound_buffer, SNDLEN);
        audio_done = 1;
        }
        else
            OSTimeDly(1);
   } while(!a320_sound_thread_exit);

    a320_sound_thread_exit=0;

    OSTaskDel(TASK_START_PRIO);
}

void a320_sound_set_volume(unsigned int involume)
{
    waveout_set_volume(involume);
}

void a320_sound_thread_mute(void)
{
   memset(a320_sound_buffer,0,SNDLEN);
}

void a320_sound_thread_stop(void)
{
   a320_sound_thread_exit=1;

   a320_sound_thread_mute();
}

void a320_sound_init(void) {
  _audio_music = waveout_open(&tempArgs);

  a320_sound_thread_mute();

  OSTaskCreate(a320_sound_thread_play, (void *) 0, (void *)&TaskStartStk[TASK_STK_SIZE - 1], TASK_START_PRIO);
}

void a320_sound_stop(void) {
    a320_sound_thread_mute();
    a320_sound_thread_stop();

    while (a320_sound_thread_exit);

    waveout_close(_audio_music);
}

void pcm_init()
{
    pcm.hz = A320SAMPLE_RATE;
	pcm.stereo = STEREO;
	pcm.len = SNDLEN>>1;
	pcm.buf = gb_sound_buffer;
	pcm.pos = 0;
    a320_sound_init();
    a320_sound_set_volume(volume);
}


int pcm_submit()
{
    if (!pcm.buf) return 0;
	if (pcm.pos < pcm.len) return 1;
	while (!audio_done)
		sys_sleep(4);
	audio_done = 0;
	pcm.pos = 0;
	return 1;
}


void pcm_close()
{
	/*if (sound) SDL_CloseAudio();*/
	a320_sound_stop();
}


/* alekmaul's scaler taken from mame4all */
void bitmap_scale(int startx, int starty, int viswidth, int visheight, int newwidth, int newheight,int pitch, uint16_t *src, uint16_t *dst) {
  unsigned int W,H,ix,iy,x,y;
  x=startx<<16;
  y=starty<<16;
  W=newwidth;
  H=newheight;
  ix=(viswidth<<16)/W;
  iy=(visheight<<16)/H;

  do
  {
    uint16_t *buffer_mem=&src[(y>>16)*320];
    W=newwidth; x=startx<<16;
    do {
      *dst++=buffer_mem[x>>16];
      x+=ix;
    } while (--W);
    dst+=pitch;
    y+=iy;
  } while (--H);
}

/* end alekmaul's scaler taken from mame4all */


void menu()
{
    bool pressed = false;
    int currentselection = 1;
    int miniscreenwidth = 140;
    int miniscreenheight = 135;
    char *cmd = NULL;
    gfx_texture* miniscreen = (gfx_texture*)malloc(sizeof(gfx_texture) + (miniscreenwidth* miniscreenheight * 2));
    miniscreen->width = miniscreenwidth;
    miniscreen->height = miniscreenheight;
    miniscreen->address = (void*)((uintptr_t)miniscreen + sizeof(gfx_texture));
    bitmap_scale(80,48,160,144,miniscreenwidth,miniscreenheight,0,(uint16_t*)fakescreen,(uint16_t*)miniscreen->address);
    char text[50];
    while (((currentselection != 1) && (currentselection != 8)) || (!pressed))
    {

        pressed = false;
        gfx_render_target_clear(gfx_color_rgb(0,0,0));

        gfx_tex_draw(320-5-miniscreenwidth,30,miniscreen);

        gfx_font_print_center(5,bigfontwhite,"DINGOO GNUBOY");

        if (currentselection == 1)
            gfx_font_print(5,25,bigfontred,"Continue");
        else
            gfx_font_print(5,25,bigfontwhite,"Continue");


        sprintf(text,"Volume %d",volume);

        if (currentselection == 2)
            gfx_font_print(5,45,bigfontred,text);
        else
            gfx_font_print(5,45,bigfontwhite,text);

        sprintf(text,"%s",useframeskip ? "Frameskip on" : "Frameskip off");

        if (currentselection == 3)
            gfx_font_print(5,65,bigfontred,text);
        else
            gfx_font_print(5,65,bigfontwhite,text);

        sprintf(text,"Save State %d",saveslot);

        if (currentselection == 4)
            gfx_font_print(5,85,bigfontred,text);
        else
            gfx_font_print(5,85,bigfontwhite,text);

        sprintf(text,"Load State %d",saveslot);

        if (currentselection == 5)
            gfx_font_print(5,105,bigfontred,text);
        else
            gfx_font_print(5,105,bigfontwhite,text);

        if (currentselection == 6)
        {
            if (fullscreen == 1)
                gfx_font_print(5,125,bigfontred,"Stretched FS");
            else
                if (fullscreen == 2)
                    gfx_font_print(5,125,bigfontred,"Aspect FS");
                else
                    gfx_font_print(5,125,bigfontred,"Native Res");
        }
        else
        {
            if (fullscreen == 1)
                gfx_font_print(5,125,bigfontwhite,"Stretched FS");
            else
                if (fullscreen == 2)
                    gfx_font_print(5,125,bigfontwhite,"Aspect FS");
                else
                    gfx_font_print(5,125,bigfontwhite,"Native Res");
        }

        sprintf(text,"%s",showfps ? "Show fps on" : "Show fps off");

        if (currentselection == 7)
            gfx_font_print(5,145,bigfontred,text);
        else
            gfx_font_print(5,145,bigfontwhite,text);


        if (currentselection == 8)
            gfx_font_print(5,165,bigfontred,"Quit");
        else
            gfx_font_print(5,165,bigfontwhite,"Quit");









        gfx_font_print_center(240-40-gfx_font_height(font),font,"Dingoo GnuBoy has been ported by joyrider");
        gfx_font_print_center(240-30-gfx_font_height(font),font,"Thanks to alekmaul for the scaler and sound example,");
        gfx_font_print_center(240-20-gfx_font_height(font),font,"Flatmush for his awesome minilib,");
        gfx_font_print_center(240-10-gfx_font_height(font),font,"Harteex for testing and the tga loading,");
        gfx_font_print_center(240-gfx_font_height(font),font,"Alf for the beautiful gameboy frames !");

        control_poll();

        if (control_check(CONTROL_DPAD_UP).pressed && control_check(CONTROL_DPAD_UP).changed)
        {
            currentselection--;
            if (currentselection == 0)
                currentselection = 8;
        }

        if (control_check(CONTROL_DPAD_DOWN).pressed && control_check(CONTROL_DPAD_DOWN).changed)
        {
            currentselection++;
            if (currentselection == 9)
                currentselection = 1;
        }

        if( ( control_check(CONTROL_BUTTON_A).pressed && control_check(CONTROL_BUTTON_A).changed) ||
            ( control_check(CONTROL_BUTTON_B).pressed && control_check(CONTROL_BUTTON_B).changed) ||
            ( control_check(CONTROL_BUTTON_START).pressed && control_check(CONTROL_BUTTON_START).changed))
                pressed = true;

         if (control_check(CONTROL_DPAD_LEFT).pressed && control_check(CONTROL_DPAD_LEFT).changed && (currentselection == 2))
        {
            volume--;
            if (volume < 0)
                volume = 0;
            a320_sound_set_volume(volume);
        }

        if (control_check(CONTROL_DPAD_RIGHT).pressed && control_check(CONTROL_DPAD_RIGHT).changed && (currentselection == 2))
        {
            volume++;
            if (volume >30)
                volume = 30;
            a320_sound_set_volume(volume);
        }

        if (control_check(CONTROL_DPAD_RIGHT).pressed && control_check(CONTROL_DPAD_RIGHT).changed && (currentselection == 3))
        {
            useframeskip = !useframeskip;
        }

        if (control_check(CONTROL_DPAD_LEFT).pressed && control_check(CONTROL_DPAD_LEFT).changed && (currentselection == 3))
        {
            useframeskip = !useframeskip;
        }

        if (control_check(CONTROL_DPAD_RIGHT).pressed && control_check(CONTROL_DPAD_RIGHT).changed && ((currentselection == 4) || (currentselection == 5)))
        {
             saveslot++;
             if (saveslot > 9)
                saveslot = 9;
            cmd = malloc(strlen("set saveslot X") +1);
            sprintf(cmd,"%s%d","set saveslot ",saveslot);
            rc_command(cmd);
            free(cmd);
        }

        if (control_check(CONTROL_DPAD_LEFT).pressed && control_check(CONTROL_DPAD_LEFT).changed && ((currentselection == 4) || (currentselection == 5)))
        {
            saveslot--;
            if (saveslot < 1)
                saveslot = 1;
            cmd = malloc(strlen("set saveslot X") +1);
            sprintf(cmd,"%s%d","set saveslot ",saveslot);
            rc_command(cmd);
            free(cmd);
        }

        if (control_check(CONTROL_DPAD_RIGHT).pressed && control_check(CONTROL_DPAD_RIGHT).changed && (currentselection == 6))
        {
            fullscreen++;
            if (fullscreen > 2)
                fullscreen = 0;
        }

        if (control_check(CONTROL_DPAD_LEFT).pressed && control_check(CONTROL_DPAD_LEFT).changed && (currentselection == 6))
        {
            fullscreen--;
            if (fullscreen < 0)
                fullscreen = 2;
        }

        if (control_check(CONTROL_DPAD_RIGHT).pressed && control_check(CONTROL_DPAD_RIGHT).changed && (currentselection == 7))
        {
            showfps = !showfps;
        }

        if (control_check(CONTROL_DPAD_LEFT).pressed && control_check(CONTROL_DPAD_LEFT).changed && (currentselection == 7))
        {
            showfps = !showfps;
        }

        if (pressed)
        {
            switch(currentselection)
            {
                case 3:
                    useframeskip = !useframeskip;
                    break;
                case 7:
                    showfps = !showfps;
                    break;
                case 6 :
                    fullscreen++;
                    if (fullscreen > 2)
                        fullscreen = 0;
                    break;
                case 4 :
                	cmd = malloc(strlen("savestate") +1);
                    sprintf(cmd,"%s","savestate");
                    rc_command(cmd);
                    free(cmd);
                    currentselection = 1;
                    break;
                case 5 :
                    cmd = malloc(strlen("loadstate") +1);
                    sprintf(cmd,"%s","loadstate");
                    rc_command(cmd);
                    free(cmd);
                    currentselection = 1;
                    break;
            }
        }

        gfx_render_target_swap();
        OSTimeDly(4);
    }
    if (currentselection == 8)
    {
        a320_sound_thread_mute();
        emuquit = 1;
    }
    free(miniscreen);
}

void ev_poll()
{
	event_t ev;
	control_poll();

    if(control_check(CONTROL_BUTTON_SELECT).pressed && control_check(CONTROL_BUTTON_START).pressed)
    {
        //fixes the select and / or start key being hang
        ev.type = EV_RELEASE;
        ev.code = 'o';
        ev_postevent(&ev);
        ev.type = EV_RELEASE;
        ev.code = 'p';
        ev_postevent(&ev);
        //end fixes the select and / or start key being hang
        a320_sound_thread_mute();
        a320_sound_paused = 1;
        menu();
        if(fullscreen == 0)
            gfx_tex_draw(0,0,frame);
        else
            gfx_render_target_clear(gfx_color_rgb(0,0,0));

        if (emuquit == 0)
            a320_sound_paused = 0;
        else
        {
            gfx_render_target_clear(gfx_color_rgb(0,0,0));
            a320_sound_paused = 1;
        }

        gfx_render_target_swap();

    }
    else
    {


        if(control_check(CONTROL_TRIGGER_RIGHT).pressed && control_check(CONTROL_TRIGGER_RIGHT).changed)
        {
            ev.type = EV_PRESS;
            ev.code = 'i';
            ev_postevent(&ev);
        }

        if(control_check(CONTROL_BUTTON_B).pressed && control_check(CONTROL_BUTTON_B).changed)
        {
            ev.type = EV_PRESS;
            ev.code = 'b';
            ev_postevent(&ev);
        }

        if(control_check(CONTROL_BUTTON_X).pressed && control_check(CONTROL_BUTTON_X).changed)
        {
            ev.type = EV_PRESS;
            ev.code = 'x';
            ev_postevent(&ev);
        }

        if(control_check(CONTROL_BUTTON_Y).pressed && control_check(CONTROL_BUTTON_Y).changed)
        {
           /* gfx_texture* temp = (gfx_texture*)malloc(sizeof(gfx_texture) + (320 * 240 * 2));
            temp->width = 320;
            temp->height = 240;
            temp->address = (void*)((uintptr_t)screen->buffer);
            gfx_tex_save_tga("A:\\game\\test.tga",temp);*/
            ev.type = EV_PRESS;
            ev.code = 'y';
            ev_postevent(&ev);
        }


        if(control_check(CONTROL_BUTTON_A).pressed && control_check(CONTROL_BUTTON_A).changed)
        {
            ev.type = EV_PRESS;
            ev.code = 'a';
            ev_postevent(&ev);
        }


        if(control_check(CONTROL_BUTTON_SELECT).pressed && control_check(CONTROL_BUTTON_SELECT).changed)
        {
            ev.type = EV_PRESS;
            ev.code = 'o';
            ev_postevent(&ev);
        }

        if(control_check(CONTROL_BUTTON_START).pressed && control_check(CONTROL_BUTTON_START).changed)
        {
            ev.type = EV_PRESS;
            ev.code = 'p';
            ev_postevent(&ev);
        }


        if(control_check(CONTROL_DPAD_DOWN).pressed && control_check(CONTROL_DPAD_DOWN).changed)
        {
            ev.type = EV_PRESS;
            ev.code = '2';
            ev_postevent(&ev);
        }


        if(control_check(CONTROL_DPAD_LEFT).pressed && control_check(CONTROL_DPAD_LEFT).changed)
        {
            ev.type = EV_PRESS;
            ev.code = '4';
            ev_postevent(&ev);
        }

        if(!control_check(CONTROL_BUTTON_B).pressed && control_check(CONTROL_BUTTON_B).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = 'b';
            ev_postevent(&ev);
        }

        if(!control_check(CONTROL_BUTTON_X).pressed && control_check(CONTROL_BUTTON_X).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = 'x';
            ev_postevent(&ev);
        }

        if(!control_check(CONTROL_BUTTON_A).pressed && control_check(CONTROL_BUTTON_A).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = 'a';
            ev_postevent(&ev);
        }


        if(!control_check(CONTROL_BUTTON_Y).pressed && control_check(CONTROL_BUTTON_Y).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = 'y';
            ev_postevent(&ev);
        }

        if(control_check(CONTROL_DPAD_RIGHT).pressed && control_check(CONTROL_DPAD_RIGHT).changed)
        {
            ev.type = EV_PRESS;
            ev.code = '6';
            ev_postevent(&ev);
        }

        if(control_check(CONTROL_DPAD_UP).pressed && control_check(CONTROL_DPAD_UP).changed)
        {
            ev.type = EV_PRESS;
            ev.code = '8';
            ev_postevent(&ev);
        }

        if(control_check(CONTROL_TRIGGER_LEFT).pressed && control_check(CONTROL_TRIGGER_LEFT).changed)
        {
            ev.type = EV_PRESS;
            ev.code = 'u';
            ev_postevent(&ev);
        }


        if(!control_check(CONTROL_BUTTON_SELECT).pressed && control_check(CONTROL_BUTTON_SELECT).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = 'o';
            ev_postevent(&ev);
        }

        if(!control_check(CONTROL_BUTTON_START).pressed && control_check(CONTROL_BUTTON_START).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = 'p';
            ev_postevent(&ev);
        }

        if(!control_check(CONTROL_DPAD_DOWN).pressed && control_check(CONTROL_DPAD_DOWN).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = '2';
            ev_postevent(&ev);
        }

        if(!control_check(CONTROL_DPAD_LEFT).pressed && control_check(CONTROL_DPAD_LEFT).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = '4';
            ev_postevent(&ev);
        }


        if(!control_check(CONTROL_DPAD_RIGHT).pressed && control_check(CONTROL_DPAD_RIGHT).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = '6';
            ev_postevent(&ev);
        }

        if(!control_check(CONTROL_DPAD_UP).pressed && control_check(CONTROL_DPAD_UP).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = '8';
            ev_postevent(&ev);
        }

        if(!control_check(CONTROL_TRIGGER_LEFT).pressed && control_check(CONTROL_TRIGGER_LEFT).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = 'u';
            ev_postevent(&ev);
        }

        if(!control_check(CONTROL_TRIGGER_RIGHT).pressed && control_check(CONTROL_TRIGGER_RIGHT).changed)
        {
            ev.type = EV_RELEASE;
            ev.code = 'i';
            ev_postevent(&ev);
        }
    }

}

void vid_setpal(int i, int r, int g, int b)
{
}

void vid_preinit()
{
    FSYS_FILE *f;
    char *cmd;
    f = fsys_fopen("A:\\GAME\\gnuboy.dat","rb");
    if(f)
    {
        fsys_fread(&fullscreen,sizeof(int),1,f);
        fsys_fread(&volume,sizeof(int),1,f);
        fsys_fread(&saveslot,sizeof(int),1,f);
        fsys_fread(&useframeskip,sizeof(bool),1,f);
        fsys_fread(&showfps,sizeof(bool),1,f);
        fsys_fclose(f);
    }
    else
    {
        fullscreen = 1;
        volume = 30;
        saveslot = 1;
        useframeskip = false;
    }
    cmd = malloc(strlen("set saveslot X") +1);
    sprintf(cmd,"%s%d","set saveslot ",saveslot);
    rc_command(cmd);
    free(cmd);
}

void vid_init(char* s,char* s2)
{

	screen = display_create(320, 240, 320, (DISPLAY_FORMAT_RGB565 | DISPLAY_BUFFER_STATIC), NULL, NULL);
	gfx_init(screen);
    loadingscreen = gfx_tex_load_tga_from_array(loading);
	gfx_tex_draw(0,0,loadingscreen);
	gfx_render_target_swap();
    sys_sleep(20);

    fakescreen = calloc(320*240, 2);
    frame = gfx_tex_load_tga(s);
    if (!frame)
        frame = gfx_tex_load_tga(s2);
    if (!frame)
        frame = gfx_tex_load_tga("A:\\game\\gnuboy.tga");
    if (!frame)
        frame = gfx_tex_load_tga_from_array(framearray);

    font = gfx_tex_load_tga_from_array(fontarray);
    bigfontwhite = gfx_tex_load_tga_from_array(bigfontwhitearray);
    bigfontred = gfx_tex_load_tga_from_array(bigfontredarray);


    control_init();


    if(fullscreen == 0)
        gfx_tex_draw(0,0,frame);
    else
        if(fullscreen == 2)
             gfx_render_target_clear(gfx_color_rgb(0,0,0));


    gfx_render_target_swap();

	fb.w = screen->width;
	fb.h = screen->height;
	fb.ptr = fakescreen;
	fb.pelsize = 2;
	fb.pitch = 640;
	fb.indexed = 0;
	fb.cc[0].r = 3;
	fb.cc[1].r = 2;
	fb.cc[2].r = 3;
	fb.cc[0].l = 11;
	fb.cc[1].l = 5;
	fb.cc[2].l = 0;
	fb.yuv = 0;
	fb.enabled = 1;
	fb.dirty = 0;


}

void vid_close()
{
    FSYS_FILE *f;
	gfx_term();
	display_delete(screen);
	fb.enabled = 0;
	control_term();
	free(fakescreen);
	free(font);
	free(bigfontred);
	free(bigfontwhite);
	free(loadingscreen);
	free(frame);
	f = fsys_fopen("A:\\GAME\\gnuboy.dat","wb");
    if(f)
    {
        fsys_fwrite(&fullscreen,sizeof(int),1,f);
        fsys_fwrite(&volume,sizeof(int),1,f);
        fsys_fwrite(&saveslot,sizeof(int),1,f);
        fsys_fwrite(&useframeskip,sizeof(bool),1,f);
        fsys_fwrite(&showfps,sizeof(bool),1,f);
        fsys_fclose(f);
    }
}

void vid_settitle(char *title)
{
}

/* my original scaler way too slow */
short getpixel(void *src,int x,int y)
{
    int pixelpos = ((y*fb.pitch) + (x * fb.pelsize));
    short tmp;
    MEMCPY(&tmp,src+pixelpos,2);
    return tmp;
}

void putpixel(void *dst,int x,int y, short color)
{
    int pixelpos = ((y*fb.pitch)+ (x * fb.pelsize));
    MEMCPY(dst+pixelpos,&color,2);
}


void ScaleToFit(void* dst,  void *src,short x,short y, short x2, short y2)
{
    long int xstep,ystep,xratio,yratio;
    short xdiff,ydiff,xcount,ycount;
    xdiff = x2-x;
    ydiff = y2-y;
    xratio=(160<<16)/(xdiff);
    yratio=(144<<16)/(ydiff);
    for(ystep=(48<<16),ycount=y; ycount<ydiff; ystep+=yratio,ycount++)
    {
        for(xstep=(80<<16),xcount=x;xcount<xdiff; xstep+=xratio,xcount++)
        {

            putpixel(dst,xcount,ycount,getpixel(src,xstep>>16,ystep>>16));
        }
    }
}
/* end my original scaler way too slow */



void vid_begin()
{
    char Text[255];
    if(emuquit == 0)
    {
        if ((!frameskip) ||  (!useframeskip))
        {
            if (fullscreen == 1)
                bitmap_scale(80,48,160,144,320,240,0,(uint16_t*)fakescreen,(uint16_t*)screen->buffer);
            else
                if (fullscreen == 2)
                    bitmap_scale(80,48,160,144,267,240,53,(uint16_t*)fakescreen,(uint16_t*)screen->buffer+25);
                else
                    bitmap_scale(80,48,160,144,160,144,160,(uint16_t*)fakescreen,(uint16_t*)screen->buffer+15440);
               // MEMCPY(screen->buffer+30880,fakescreen+30880,184640);
            //ScaleToFit(screen->buffer,fakescreen,0,0,320,240);
            //MEMCPY(screen->buffer,fakescreen,320*240*2);

            if(showfps)
            {
                sprintf(Text,"%d",fps);
                gfx_rect_fill_draw(0,0,gfx_font_width(font,Text),gfx_font_height(font),0);
                gfx_font_print(0,0,font,Text);
                framecounter++;
                if (clock() - time1 > CLOCKS_PER_SEC)
                {
                    fps=framecounter;
                    framecounter=0;
                    time1 = clock();
                }
            }
            gfx_render_target_swap();
        }
    }
    frameskip = !frameskip;

}

void vid_end()
{

}



